"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from tumourkit.utils.preprocessing import *

def test_true():
    assert True



import argparse
from argparse import Namespace
import logging
from logging import Logger
import os
from typing import List
import pandas as pd


class WrongConfigurationError(Exception):
    pass


def check_void(path: str) -> bool:
    """Tells if a directory if void or if file exists. True if void, False otherwise."""
    if os.path.isdir(path):
        files = os.listdir(path)
        return len(files) == 0
    return not os.path.exists(path)


def check_same_num(dir_list: List[str]) -> bool:
    """Tells if all the folders in the list contain the same number of files."""
    num_files = [len([os.listdir(dir) for dir in dir_list])]
    if len(num_files) == 0:
        return True
    num_file = num_files[0]
    for element in num_files:
        if element != num_file:
            return False
    return True


def run_preprocessing(args: Namespace, logger: Logger) -> None:
    logger.info('Checking training pipeline was run correctly.')
    check_list = [
        os.path.join(args.root_dir, 'gnn_logs', 'gnn_results.csv'),
        os.path.join(args.root_dir, 'weights', 'classification', 'gnn', 'confs'),
        os.path.join(args.root_dir, 'weights', 'classification', 'gnn', 'normalizers'),
        os.path.join(args.root_dir, 'weights', 'classification', 'gnn', 'weights'),
        os.path.join(args.root_dir, 'data', 'train', 'graphs', 'preds'),
        os.path.join(args.root_dir, 'data', 'validation', 'graphs', 'preds'),
        os.path.join(args.root_dir, 'data', 'test', 'graphs', 'preds'),
    ]
    for element in check_list:
        if check_void(element):
            raise WrongConfigurationError(f'The following folder / file was not found: {element}. Try running the training pipeline again.')
        
    dir_list = [os.path.join(args.root_dir, 'weights', 'classification', 'gnn', dir) for dir in ['confs', 'normalizers', 'weights']]
    if not check_same_num(dir_list):
        raise WrongConfigurationError('Graph weights folders contain different number of files. Try running the training pipeline again')
    
    results = pd.read_csv(os.path.join(args.root_dir, 'gnn_logs', 'gnn_results.csv'))
    if 'F1 Score' in results.columns and args.num_classes != 2:
        raise WrongConfigurationError('You have trained the GNN models for binary problem but are asking to run research for multiclass.')
    elif 'Micro F1' in results.columns and args.num_classes == 2:
        raise WrongConfigurationError('You have trained the GNN models for multiclass problem but are asking to run research for binary.')

    return


def run_models(args: Namespace, logger: Logger) -> None:
    return


def run_evaluation(args: Namespace, logger: Logger) -> None:
    return


def run_visualizations(args: Namespace, logger: Logger) -> None:
    return


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--root-dir', type=str, help='Internal data and weights directory.', default='./.internals/')
    parser.add_argument('--output-dir', type=str, help='Folder where to save all the results.')
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    return parser


def main():
    parser = _create_parser()
    args = parser.parse_args()
    
    logger = logging.getLogger('research_pipe')
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    if args.num_classes != 2:
        raise NotImplementedError('This pipe only supports the binary case.')

    run_preprocessing(args, logger)
    run_models(args, logger)
    run_evaluation(args, logger)
    run_visualizations(args, logger)



"""
Script to train xgboost models.
Right now only supports classification over nodes, without edges.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import argparse
from argparse import Namespace
from .read_nodes import create_node_splits
from xgboost import XGBClassifier
import os
from sklearn.metrics import f1_score, accuracy_score, roc_auc_score
from sklearn.model_selection import train_test_split
import numpy as np
from typing import Dict, Any, Tuple
from sklearn.model_selection import StratifiedKFold
import pandas as pd
from concurrent.futures import ThreadPoolExecutor

FILE_DIR = os.path.dirname(os.path.abspath(__file__))




def train(
        conf: Dict[str, Any], 
        X_tr: np.ndarray, 
        y_tr: np.ndarray, 
        val_size: float,
        seed: int
        ) -> XGBClassifier:
    X_tr, X_val, y_tr, y_val = train_test_split(
        X_tr, y_tr, test_size=val_size, stratify=y_tr, random_state=seed
    )
    model = XGBClassifier(
        n_estimators=conf['n_estimators'],
        learning_rate=conf['learning_rate'], # 0.05
        max_depth=conf['max_depth'],
        colsample_bytree=conf['colsample_bytree'],
        eval_metric='logloss',
        early_stopping_rounds=10
    )
    model.fit(
        X_tr, y_tr,
        eval_set=[(X_val, y_val)], verbose=False
    )
    return model

def evaluate(
        model: XGBClassifier, 
        X_val: np.ndarray,
        y_val: np.ndarray
        ) -> Tuple[float, float, float]:
    preds = model.predict(X_val)
    probs = model.predict_proba(X_val)[:,1]
    f1 = f1_score(y_val, preds)
    acc = accuracy_score(y_val, preds)
    auc = roc_auc_score(y_val, probs)
    return f1, acc, auc

def save(metrics: Dict[str,Tuple[float,float,float]], path: str) -> None:
    metrics.to_csv(path, index=False)

def cross_validate(args, conf, skf, X, y):
    f1_mean, acc_mean, auc_mean = 0, 0, 0
    for train_index, val_index in skf.split(X, y):
        X_train, X_cval = X[train_index], X[val_index]
        y_train, y_cval = y[train_index], y[val_index]
        ## Train
        model = train(conf, X_train, y_train, args.val_size, args.seed)
        ## Save metrics
        f1, acc, auc = evaluate(model, X_cval, y_cval)
        f1_mean += f1; acc_mean += acc; auc_mean += auc
    f1_mean /= args.cv_folds; acc_mean /= args.cv_folds; auc_mean /= args.cv_folds
    tmp = pd.DataFrame(
        [(*conf.values(), f1_mean, acc_mean, auc_mean)], 
        columns=[
            'n_estimators', 'learning_rate', 'max_depth', 'colsample_bytree',
            'f1', 'accuracy', 'auc'
            ]
    )
    return tmp

def create_confs():
    confs = [{'n_estimators': n, 'learning_rate': l, 'max_depth': d, 'colsample_bytree': c}
            for n in [100, 500] 
            for l in [0.05, 0.005] 
            for d in [4, 8, 16] 
            for c in [0, 0.5]]
    return confs


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--graph-dir', type=str, required=True,
                        help='Folder containing .graph.csv files.')
    parser.add_argument('--val-size', type=float, default=0.2,
                        help='Validation size used for early stopping. Default: 0.2')
    parser.add_argument('--seed', type=int, default=None,
                        help='Seed for random split. Default: None')
    parser.add_argument('--num-workers', type=int, default=1, 
                        help='Number of processors to use. Default: 1.')
    parser.add_argument('--cv-folds', type=int, default=10, 
                        help='Number of CV folds. Default: 10.')
    parser.add_argument('--save-name', type=str, required=True,
                        help='Name to save the result, without file type.')
    return parser


def main_with_args(args: Namespace):
    GRAPH_DIR = args.graph_dir

    X_train, X_val, X_test, y_train, y_val, y_test = \
        create_node_splits(
            GRAPH_DIR, 0.2, 0.2, 0, 'total'
        )

    X = np.vstack((X_train, X_val, X_test))
    y = np.hstack((y_train, y_val, y_test))
    y = np.array(y, dtype=np.int32)
    skf = StratifiedKFold(n_splits=args.cv_folds)
    skf.get_n_splits(X, y)

    metrics = pd.DataFrame(
        {}, 
        columns=[
            'n_estimators', 'learning_rate', 'max_depth', 'colsample_bytree',
            'f1', 'accuracy', 'auc'
        ]
    )
    confs = create_confs()
    with ThreadPoolExecutor(max_workers=args.num_workers) as executor:
        futures = []
        for conf in confs:
            future = executor.submit(cross_validate, args, conf, skf, X, y)
            futures.append(future)
        for k, future in enumerate(futures):
            print('Configuration {:2}/{:2}'.format(k, len(confs)), end='\r')
            tmp = future.result()
            metrics = pd.concat((metrics, tmp))
            save(metrics, args.save_name + '.csv')
    for k, conf in enumerate(confs):
        print('Configuration {:2}/{:2}'.format(k, len(confs)), end='\r')
        tmp = cross_validate(args, conf, skf, X, y)
        metrics = pd.concat((metrics, tmp))
        save(metrics, args.save_name + '.csv')
    save(metrics, args.save_name + '.csv')


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)
    
    


"""
Module with utility functions for the classification module.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import numpy as np
from sklearn.preprocessing import StandardScaler
from typing import Tuple, List
from ..classification.read_nodes import read_all_nodes

def normalize(X_train: np.ndarray, X_val: np.ndarray, X_test: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Normalizes the input data using the mean and standard deviation of the training dataset.

    :param X_train: The input training data with shape (n_samples, n_features).
    :type X_train: numpy.ndarray
    :param X_val: The input validation data with shape (n_samples, n_features).
    :type X_val: numpy.ndarray
    :param X_test: The input test data with shape (n_samples, n_features).
    :type X_test: numpy.ndarray
    :return: A tuple containing three numpy.ndarrays:
             - The normalized X_train with shape (n_samples, n_features).
             - The normalized X_val with shape (n_samples, n_features).
             - The normalized X_test with shape (n_samples, n_features).
    :rtype: tuple(numpy.ndarray, numpy.ndarray, numpy.ndarray)
    """
    sc = StandardScaler()
    sc.fit(X_train)
    return sc.transform(X_train), sc.transform(X_val), sc.transform(X_test)

def fit_column_normalizer(node_dir: str, names: List[str]) -> StandardScaler:
    """
    Fits a StandardScaler object to the input data and returns it.

    :param node_dir: The directory containing node data files.
    :type node_dir: str
    :param names: A list of node names to read from the node directory.
    :type names: List[str]
    :return: A StandardScaler object fitted to the input data.
    :rtype: StandardScaler
    """
    X, y = read_all_nodes(node_dir, [x+'.nodes.csv' for x in names])
    sc = StandardScaler()
    sc.fit(X)
    return sc